"""mokeidb

By deuxglaces
"""
import pathlib

import mokeidb.mongodb
from mokeidb.enum import EnumMixin
from mokeidb.exceptions import DatabaseError, UniqueConflict, NoDatabase
from mokeidb.decorators import collection, database
from mokeidb.generics import Key, Unique, NotNull
from mokeidb.mongodb import MongoDbModel, mongodb_config
from mokeidb.sqlite import SqliteDbModel

__version__ = '0.1.6'



import collections.abc
import datetime
import sqlite3

import mokeidb.dbmodel

_DATA_CAST: dict[type, tuple[collections.abc.Callable, collections.abc.Callable]] = {
    datetime.date: (datetime.date.isoformat, datetime.date.fromisoformat),
    datetime.time: (datetime.time.isoformat, datetime.time.fromisoformat),
    datetime.datetime: (datetime.datetime.isoformat, datetime.datetime.fromisoformat),
}


def register_type(type_, store_fn: collections.abc.Callable, retrive_fn: collections.abc.Callable):
    _DATA_CAST[type_] = (store_fn, retrive_fn)


class SqliteDbModel(mokeidb.dbmodel.DbModel):
    _CONNECTION: sqlite3.Connection
    _CURSOR: sqlite3.Cursor
    _TABLE_NAME: str

    def __init__(self, **kwargs):
        super().__init__()
        for k in self.__class__.__annotations__:
            self.__dict__[k] = kwargs[k] if k in kwargs else self.__class__._get_default(k)

    def __init_subclass__(cls, **kwargs):
        try:
            database = kwargs['database']
        except KeyError:
            raise SyntaxError('SqliteDbModel subclasses must specify database keyword arg in definition')
        try:
            cls._CONNECTION = sqlite3.connect(database)
        except Exception as e:
            print(e)
        cls._CURSOR = cls._CONNECTION.cursor()

        # database tables are, by default, the name of the subclass
        cls._TABLE_NAME = kwargs.get('table', cls.__name__)
        cls._init_table()

    @classmethod
    def _init_table(cls):
        pass

    @classmethod
    def find(cls, *args, **kwargs):
        pass

    @classmethod
    def find_one(cls, *args, **kwargs):
        pass

    @classmethod
    def all(cls):
        pass

    @classmethod
    def new(cls, *args, **kwargs):
        pass

    def update(self, **kwargs):
        pass

    def delete(self):
        pass


if __name__ == '__main__':
    register_type(str, str, str)



import mokeidb.mongodb


def database(database_or_name):
    def decorator(cls):
        if hasattr(cls, '__mro__') and mokeidb.mongodb.MongoDbModel in cls.__mro__:
            cls.set_database(database_or_name)
        return cls

    return decorator


def collection(collection_or_name):
    def decorator(cls):
        if hasattr(cls, '__mro__') and mokeidb.mongodb.MongoDbModel in cls.__mro__:
            cls.set_collection(collection_or_name)
        return cls

    return decorator



import dataclasses


@dataclasses.dataclass
class Field:
    name: str
    type: type
    key: bool
    unique: bool
    notnull: bool

    @property
    def unique_or_key(self):
        return self.unique or self.key



class DatabaseError(Exception):
    pass


class UniqueConflict(DatabaseError, ValueError):
    pass


class NoDatabase(DatabaseError, ValueError):
    pass


class NoCollection(DatabaseError, ValueError):
    pass


class MultipleKeys(DatabaseError):
    pass



import functools

import mokeidb.mongodb


def _store_enum(value):
    if hasattr(value, 'value'):
        return value.value
    raise AttributeError('EnumMixin subclasses must also subclass enum.Enum')


def _retrieve_enum(cls, value):
    if hasattr(cls, '_value2member_map_'):
        return cls._value2member_map_.get(value)
    raise AttributeError('DbStoreableEnum subclasses must also subclass enum.Enum')


class EnumMixin:
    def __init_subclass__(cls, **kwargs):
        mokeidb.mongodb.register_type(cls, _store_enum, functools.partial(_retrieve_enum, cls))



import datetime
import dataclasses
import collections.abc
import decimal
import enum
import json
import pathlib
from typing import Any, Type, TypeVar, get_args, get_origin, Generic

_T = TypeVar('_T')


def dec_to_str(dec: decimal.Decimal, min_dp: int = 0) -> str:
    """Custom function to convert decimal.Decimal to str in a consistant format with a
    min number of decimal places"""
    if dec == decimal.Decimal('0'):
        # special case to deal with zero
        sign = 0
        digits = ['0'] + ['0'] * min_dp
        exponent = - min_dp
    else:
        sign, digits_tup, exponent = dec.as_tuple()
        digits = [str(digit) for digit in digits_tup]
        while digits[-1] == '0' and len(digits) > 1:
            digits.pop()
            exponent += 1
        add_zero_after = max(min_dp + exponent, 0)
        exponent -= add_zero_after
        add_zero_before = abs(min(exponent + len(digits) + add_zero_after - 1, 0))
        digits = ['0'] * add_zero_before + digits + ['0'] * add_zero_after
    if exponent < 0:
        return ('-' if sign else '') + ''.join(digits[:exponent]) + '.' + ''.join(digits[exponent:])
    return ('-' if sign else '') + ''.join(digits) + '0' * exponent


class Serializer:
    def __init__(self):
        self._DATA_CAST: dict[type, tuple[collections.abc.Callable, collections.abc.Callable]] = {
            datetime.date: (datetime.date.isoformat, datetime.date.fromisoformat),
            datetime.time: (datetime.time.isoformat, datetime.time.fromisoformat),
            datetime.datetime: (datetime.datetime.isoformat, datetime.datetime.fromisoformat),
            decimal.Decimal: (dec_to_str, decimal.Decimal),
        }

    def register_type(self, type_, store_fn: collections.abc.Callable, retrive_fn: collections.abc.Callable):
        self._DATA_CAST[type_] = (store_fn, retrive_fn)

    def to_str(self, obj: Any) -> str:
        return json.dumps(self._ser(obj))

    def to_dict(self, obj: Any) -> dict:
        return self._ser(obj)

    def to_file(self, obj: Any, filepath: str | pathlib.Path) -> None:
        file_data = self.to_str(obj)
        with open(filepath, mode='w') as output_file:
            output_file.write(file_data)

    def from_str(self, cls: Type[_T], json_str: str) -> _T:
        return self._deser(cls, json.loads(json_str))

    def from_dict(self, cls: Type[_T], json_dict: dict) -> _T:
        return self._deser(cls, json_dict)

    def from_file(self, cls: Type[_T], filepath: str | pathlib.Path) -> _T:
        with open(filepath) as input_file:
            return self.from_str(cls, input_file.read())

    def _ser(self, v):
        if v is None:
            return None
        if dataclasses.is_dataclass(v.__class__):
            return self._ser(v.__dict__)
        if isinstance(v, enum.Enum):
            return v.value
        if v.__class__ in self._DATA_CAST:
            return self._DATA_CAST[v.__class__][0](v)
        if any(isinstance(v, c) for c in (list, tuple, set)):
            return [self._ser(item) for item in v]
        if isinstance(v, dict):
            return {self._ser(k): self._ser(va) for k, va in v.items()}
        return v

    def _deser(self, c, v):
        if dataclasses.is_dataclass(c):
            return None if v is None else c(**{
                k: self._deser(c.__annotations__[k], v.get(k)) for k in v
            })
        if c in self._DATA_CAST:
            return self._DATA_CAST[c][1](v)
        if origin := get_origin(c):
            if issubclass(c.__origin__, Generic):
                if args := get_args(c):
                    return self._deser(args[0], v)
            if origin is list:
                # type hint must be of form list[str] etc. with single type of value
                return [self._deser(c.__args__[0], item) for item in v]
            if origin is set:
                # type hint must be of form set[str] etc. with single type of value
                return {self._deser(c.__args__[0], item) for item in v}
            if origin is tuple:
                # type hint must be of form tuple[str, datetime.date, int]
                # etc. with type for each of the FIXED no. of items
                return tuple(self._deser(item_cls, item) for item_cls, item in zip(c.__args__, v))
            if origin is collections.defaultdict:
                # type hint must be of form collections.defaultdict[str, int] etc. with types of key and value mentioned
                # only works for defaultdict whose factory is the second type arg
                return collections.defaultdict(c.__args__[1], {
                    self._deser(c.__args__[0], k): self._deser(c.__args__[1], v) for k, v in
                    v.items()
                })
            if origin is dict:
                # type hint must be of form dict[str, int] etc. with types of key and value mentioned
                return {
                    self._deser(c.__args__[0], k): self._deser(c.__args__[1], v) for k, v in
                    v.items()
                }
        if issubclass(c, enum.Enum):
            return c(v)
        return v



import datetime
import unittest

import pymongo

import mokeidb.mongodb


class TestMongoDb(unittest.TestCase):
    def tearDown(self) -> None:
        with pymongo.MongoClient() as mongo_client:
            mongo_client.drop_database('MokeiDbTest')
            mongo_client.drop_database('tests')

    def test_001_new(self):
        @mokeidb.database('MokeiDbTest')
        class Employee(mokeidb.mongodb.MongoDbModel):
            employee_id: mokeidb.Unique[str]
            name: str
            dob: datetime.date

        @mokeidb.database('MokeiDbTest')
        class Department(mokeidb.mongodb.MongoDbModel):
            department_code: mokeidb.Unique[str]
            name: mokeidb.Unique[str]
            employees: list[Employee]

        self.assertEqual(0, len(Employee.all()))
        self.assertEqual(0, len(Department.all()))
        test_employee = Employee(
            employee_id='NewEmployeeId',
            name='Deux Glaces',
            dob=datetime.date(1999, 12, 31),
        )

        test_department = Department(
            department_code='Weirdoes',
            name='Pythonistas!',
            employees=[test_employee],
        )

        self.assertEqual('NewEmployeeId', test_employee.employee_id)
        self.assertEqual('Deux Glaces', test_employee.name)
        self.assertEqual(datetime.date(1999, 12, 31), test_employee.dob)

        self.assertEqual('Weirdoes', test_department.department_code)
        self.assertEqual('Pythonistas!', test_department.name)
        self.assertEqual([test_employee], test_department.employees)
        self.assertIs(test_employee, test_department.employees[0])

        self.assertEqual(1, len(Employee.all()))
        retrieved_employee = Employee.find_one()
        self.assertIs(test_employee, retrieved_employee)
        self.assertEqual(datetime.date(1999, 12, 31), retrieved_employee.dob)

        del test_employee
        del test_department
        retrieved_employee = Employee.find_one()
        retrieved_dept = Department.find_one()
        self.assertEqual(datetime.date(1999, 12, 31), retrieved_employee.dob)
        self.assertIs(retrieved_dept.employees[0], retrieved_employee)

    def test_002_update_and_save(self):
        @mokeidb.database('MokeiDbTest')
        class Employee(mokeidb.mongodb.MongoDbModel):
            employee_id: mokeidb.Unique[str]
            name: mokeidb.NotNull[str]
            dob: datetime.date

        @mokeidb.database('MokeiDbTest')
        class Department(mokeidb.mongodb.MongoDbModel):
            department_code: mokeidb.Unique[str]
            name: mokeidb.Unique[str]
            employees: list[Employee]

        self.assertEqual(0, len(Employee.all()))
        self.assertEqual(0, len(Department.all()))
        test_employee = Employee(
            employee_id='NewEmployeeId',
            name='Deux Glaces',
            dob=datetime.date(1999, 12, 31),
        )

        self.assertEqual('NewEmployeeId', test_employee.employee_id)
        self.assertEqual('Deux Glaces', test_employee.name)
        self.assertEqual(datetime.date(1999, 12, 31), test_employee.dob)

        self.assertEqual(1, len(Employee.all()))
        retrieved_employee = Employee.find_one()
        self.assertIs(test_employee, retrieved_employee)
        self.assertEqual(datetime.date(1999, 12, 31), retrieved_employee.dob)

        # update changes a single object, and all references will be updated at once
        retrieved_employee.update(name='Dooo Glaces')
        self.assertEqual('Dooo Glaces', test_employee.name)
        self.assertEqual('Dooo Glaces', retrieved_employee.name)

        # update with no args does nothing
        retrieved_employee.update()
        del test_employee
        retrieved_employee = Employee.find_one()
        self.assertEqual('Dooo Glaces', retrieved_employee.name)

        # save updates database after changing instance
        retrieved_employee.name = 'Deux Glaces'
        retrieved_employee.save()
        self.assertEqual('Deux Glaces', retrieved_employee.name)
        del retrieved_employee
        retrieved_employee = Employee.find_one()
        self.assertEqual('Deux Glaces', retrieved_employee.name)

        # save updates kwargs too
        retrieved_employee.name = 'Pi Thagoras'
        retrieved_employee.save(dob=datetime.date(2020, 3, 14))
        del retrieved_employee
        retrieved_employee = Employee.find_one()
        self.assertEqual('Pi Thagoras', retrieved_employee.name)
        self.assertEqual(datetime.date(2020, 3, 14), retrieved_employee.dob)

        # update only updates the kwargs, not the other fields
        retrieved_employee.name = 'Don\'t name your child this'
        retrieved_employee.update(dob=datetime.date(2020, 12, 25))
        del retrieved_employee
        retrieved_employee = Employee.find_one()
        self.assertEqual('Pi Thagoras', retrieved_employee.name)
        self.assertEqual(datetime.date(2020, 12, 25), retrieved_employee.dob)

        # update a unique field to a value that is already present in a different document
        new_employee = Employee(
            employee_id='StanEmployeeId',
            name='Old Timer',
            dob=datetime.date(1921, 1, 1)
        )
        self.assertRaises(
            mokeidb.exceptions.UniqueConflict,
            new_employee.update,
            employee_id='NewEmployeeId',
        )
        # but different value doesn't raise the exception
        new_employee.update(employee_id='DifferentEmployeeId')
        new_employee.delete()

    def test_901_test_config(self):
        mokeidb.mongodb_config(database='MokeiDbTest')

        class Employee(mokeidb.mongodb.MongoDbModel):
            employee_id: mokeidb.Unique[str]
            name: str
            dob: datetime.date

        class Department(mokeidb.mongodb.MongoDbModel):
            department_code: mokeidb.Unique[str]
            name: mokeidb.Unique[str]
            employees: list[Employee]

        self.assertEqual(0, len(Employee.all()))
        self.assertEqual(0, len(Department.all()))
        test_employee = Employee(
            employee_id='NewEmployeeId',
            name='Deux Glaces',
            dob=datetime.date(1999, 12, 31),
        )

        test_department = Department(
            department_code='Weirdoes',
            name='Pythonistas!',
            employees=[test_employee],
        )

        self.assertEqual('NewEmployeeId', test_employee.employee_id)
        self.assertEqual('Deux Glaces', test_employee.name)
        self.assertEqual(datetime.date(1999, 12, 31), test_employee.dob)

        self.assertEqual('Weirdoes', test_department.department_code)
        self.assertEqual('Pythonistas!', test_department.name)
        self.assertEqual([test_employee], test_department.employees)
        self.assertIs(test_employee, test_department.employees[0])

        self.assertEqual(1, len(Employee.all()))
        retrieved_employee = Employee.find_one()
        self.assertIs(test_employee, retrieved_employee)
        self.assertEqual(datetime.date(1999, 12, 31), retrieved_employee.dob)

        del test_employee
        del test_department
        retrieved_employee = Employee.find_one()
        retrieved_dept = Department.find_one()
        self.assertEqual(datetime.date(1999, 12, 31), retrieved_employee.dob)
        self.assertIs(retrieved_dept.employees[0], retrieved_employee)



import collections.abc

import mokeidb.field
import mokeidb.generics


class DbModel:
    def __new__(cls, *args, **kwargs):
        if '___internal___' in kwargs:
            kwargs.pop('___internal___')
            return super().__new__(cls)
        return cls.new(*args, **kwargs)

    @classmethod
    def _get_field(cls, fieldname: str, key=False, unique=False, notnull=False, anno=None) -> mokeidb.field.Field:
        anno = anno or cls.__annotations__[fieldname]
        if hasattr(anno, '__origin__'):
            if anno.__origin__ is mokeidb.generics.Key:
                return cls._get_field(fieldname, key=True, unique=unique, notnull=notnull, anno=anno.__args__[0])
            if anno.__origin__ is mokeidb.generics.Unique:
                return cls._get_field(fieldname, key=key, unique=True, notnull=notnull, anno=anno.__args__[0])
            if anno.__origin__ is mokeidb.generics.NotNull:
                return cls._get_field(fieldname, key=key, unique=key, notnull=True, anno=anno.__args__[0])
            if anno.__origin__ in {set, list, dict, tuple}:
                return mokeidb.field.Field(fieldname, anno.__origin__, key, unique, notnull)

        return mokeidb.field.Field(fieldname, anno, key, unique, notnull)

    @classmethod
    def _get_default(cls, fieldname: str):
        default = cls.__dict__.get(fieldname)
        if isinstance(default, collections.abc.Callable):
            return default()
        if isinstance(default, collections.abc.Iterator):
            return next(default)
        t = cls._get_field(fieldname).type
        if t in {list, set, dict, tuple}:
            return t()
        return default

    @classmethod
    def find(cls, *args, **kwargs):
        pass

    @classmethod
    def find_one(cls, *args, **kwargs):
        pass

    @classmethod
    def all(cls):
        return cls.find()

    @classmethod
    def new(cls, *args, **kwargs):
        pass

    @classmethod
    def get_or_create(cls, *args, **kwargs):
        pass

    def update(self, **kwargs):
        """Update the database ONLY with values from kwargs.  Any other updated values will not be saved.
        If you need
        :param kwargs:
        :return:
        """
        pass

    def delete(self):
        pass

    def save(self, **kwargs):
        """Save the current state of the instance to the database, after updating any fields with values
        from kwargs
        :param kwargs:
        :return:
        """
        update_vals = {
            k: v for k, v in self.__dict__.items() if k in self.__annotations__
        }
        update_vals.update(kwargs)
        return self.update(**update_vals)

    def _db_store_(self):
        pass

    @classmethod
    def _db_retrieve_(cls, stored_val):
        pass



import datetime
import unittest

from mokeidb.serializer import Serializer


class TestSerializer(unittest.TestCase):
    def test_serialize(self):
        serializer = Serializer()
        self.assertEqual('2021-12-25', serializer.to_dict(datetime.date(2021, 12, 25)))
        self.assertEqual(datetime.date(2021, 12, 25), serializer.from_dict(datetime.date, '2021-12-25'))



import collections.abc
import inspect
import os
import pathlib
import weakref
from typing import Self, Optional, Callable

import bson
import pymongo.database
import pymongo.errors

from .dbmodel import DbModel
from .exceptions import MultipleKeys, NoCollection, NoDatabase, UniqueConflict
from .generics import Key, Unique
from .serializer import Serializer


def register_type(type_, store_fn: collections.abc.Callable, retrive_fn: collections.abc.Callable):
    MongoDbModel.register_type(type_, store_fn, retrive_fn)


# noinspection PyProtectedMember
def mongodb_config(
        *,
        connection_uri: str = None,
        client: pymongo.MongoClient = None,
        database: [str, pymongo.database.Database] = None,
        host=None,
        port=None,
        document_class=dict,
        tz_aware=None,
        connect=None,
        type_registry=None,
):
    """Single time config for MongoDb to set up a particular client, database etc. for all database models,
    avoiding repeat decorators etc.  This must be called before any class definitions subclassing MongoDbModel.

    The following code is equivalent:
    Snippet 1:
        @database('MyDB')
        class FirstModel(MongoDbModel):
            pass

        @database('MyDB')
        class FirstModel(MongoDbMOdel):
            pass

    Snippet 2:
        mongodb_config(database='MyDB')

        class FirstModel(MongoDbModel):
            pass

        class FirstModel(MongoDbMOdel):
            pass
    """
    # set the default client, and default database on the MongoDbModel class, NOT on subclasses directly
    if client:
        MongoDbModel._DEFAULT_CLIENT = client
    elif connection_uri:
        MongoDbModel._DEFAULT_CLIENT = pymongo.MongoClient(connection_uri, tz_aware=tz_aware, connect=connect,
                                                           type_registry=type_registry)
    elif (
            any(kwarg is not None for kwarg in [host, port, tz_aware, connect, type_registry]) or
            document_class is not dict
    ):
        client = pymongo.MongoClient(host, port, document_class, tz_aware, connect, type_registry)
        MongoDbModel._DEFAULT_CLIENT = client

    if isinstance(database, str):
        MongoDbModel._Defaults.DATABASE_NAME = database
    elif isinstance(database, pymongo.database.Database):
        MongoDbModel._Defaults.DATABASE_NAME = database.name


class MongoDbModel(DbModel):
    # MongoDbModel Default attributes
    class _Defaults:
        CLIENT: pymongo.MongoClient = None
        DATABASE_NAME: str = pathlib.Path(os.getcwd()).name
        SERIALIZER: Serializer = Serializer()

    # Subclass attributes (set in __init_subclass__)
    _CLIENT: Optional[pymongo.MongoClient]
    _DATABASE: Optional[pymongo.database.Database]
    _DATABASE_NAME: Optional[str]
    _COLLECTION: Optional[pymongo.database.Collection]
    _COLLECTION_NAME: Optional[str]
    _SERIALIZER: Optional[Serializer]
    _WEAKREFS: dict[bson.ObjectId, weakref.ref]
    _UNIQUE_FIELDS: list[str]
    _KEY_FIELD: str = '_id'

    # Instance attributes (set in __init__)
    _id: bson.ObjectId

    def __init__(self, *args, **kwargs):
        self.__class__.lazy_init_subclass()
        # initialize with default values where specified, None otherwise
        for key in self.__annotations__:
            default = self.__class__.__dict__.get(key)
            if isinstance(default, Callable):
                default = default()
            self.__dict__[key] = default

        values = {k: v for k, v in zip(self.__annotations__, args)}
        values.update(kwargs)
        self.__dict__.update(values)
        self.__class__._WEAKREFS[self._id] = weakref.ref(self)

    def __init_subclass__(cls, **kwargs):
        cls._SUBCLASS_INITIALIZED = False
        cls._CLIENT = None
        cls._DATABASE = None
        cls._DATABASE_NAME = None
        cls._COLLECTION = None
        cls._COLLECTION_NAME = None
        cls._SERIALIZER = None
        cls._WEAKREFS = {}
        cls._UNIQUE_FIELDS = []

        for anno, val in cls.__annotations__.items():
            if hasattr(val, '__origin__') and val.__origin__ is Key:
                if cls._KEY_FIELD != '_id':
                    raise MultipleKeys(f'Too many Key fields in model "{cls.__name__}". Use Unique'
                                       f' for multiple unique fields.')
                cls._KEY_FIELD = anno
                cls._UNIQUE_FIELDS.append(anno)
            if hasattr(val, '__origin__') and val.__origin__ is Unique:
                cls._UNIQUE_FIELDS.append(anno)
        cls.__init__.__signature__ = inspect.Signature(
            parameters=[inspect.Parameter('self', inspect.Parameter.POSITIONAL_ONLY, annotation=cls)] + [
                inspect.Parameter(k, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=v)
                for k, v in cls.__annotations__.items()
            ],
            return_annotation=None,
        )
        register_type(cls, cls._db_store_, cls._db_retrieve_)

    @classmethod
    def lazy_init_subclass(cls) -> None:
        if not cls._SUBCLASS_INITIALIZED:
            if cls._CLIENT is None:
                if cls._Defaults.CLIENT is None:
                    cls._Defaults.CLIENT = pymongo.MongoClient()
                cls._CLIENT = cls._Defaults.CLIENT
            if cls._DATABASE is None:
                if cls._DATABASE_NAME:
                    cls._DATABASE = cls._CLIENT.get_database(cls._DATABASE_NAME)
                else:
                    cls._DATABASE = cls._CLIENT.get_database(cls._Defaults.DATABASE_NAME)
            if cls._COLLECTION is None:
                if not cls._COLLECTION_NAME:
                    cls._COLLECTION_NAME = cls.__name__
                cls._COLLECTION = cls._DATABASE.get_collection(cls._COLLECTION_NAME)
                for unique_field in cls._UNIQUE_FIELDS:
                    cls._COLLECTION.create_index(unique_field, unique=True)
                cls._COLLECTION_INITIALIZED = True
            if cls._SERIALIZER is None:
                cls._SERIALIZER = cls._Defaults.SERIALIZER
            cls._SUBCLASS_INITIALIZED = True

    @classmethod
    def register_type(cls, type_, store_fn, retrive_fn):
        if cls is MongoDbModel:
            cls._Defaults.SERIALIZER.register_type(type_, store_fn, retrive_fn)
        else:
            if cls._SERIALIZER is None:
                cls._SERIALIZER = cls._Defaults.SERIALIZER
            cls._SERIALIZER.register_type(type_, store_fn, retrive_fn)

    @classmethod
    def set_database(cls, database: [str, pymongo.database.Database]):
        if isinstance(database, pymongo.database.Database):
            cls._DATABASE = database
            return
        if isinstance(database, str):
            cls._DATABASE_NAME = database
            return
        raise NoDatabase(f'Unable to set database for {cls}')

    @classmethod
    def get_database(cls):
        return cls._DATABASE

    @classmethod
    def _set_client(cls, client: [str, pymongo.MongoClient]):
        cls._CLIENT = client

    @classmethod
    def _get_client(cls):
        return cls._CLIENT

    @classmethod
    def set_collection(cls, collection: [str, pymongo.database.Collection]):
        if isinstance(collection, pymongo.database.Collection):
            cls._COLLECTION = collection
            return
        elif isinstance(collection, str):
            if (database := cls.get_database()) is not None:
                cls._COLLECTION = database.get_collection(collection)
                return
        raise NoDatabase(f'Unable to set collection to {collection} for {cls} as there is no specified database')

    @classmethod
    def _init_collection(cls, collection: pymongo.database.Collection):
        for k in cls.__annotations__:
            if cls._get_field(k).unique_or_key:
                collection.create_index(k, unique=True)
        cls._COLLECTION_INITIALIZED = True

    @classmethod
    def get_collection(cls):
        collection = None
        if cls._COLLECTION is not None:
            collection = cls._COLLECTION
        database = cls.get_database()
        if database is not None:
            collection = database.get_collection(cls.__name__)
        if collection is None:
            raise NoCollection(f'{cls} has no specified collection, or database for '
                               f'generic collection')
        if not cls._COLLECTION_INITIALIZED:
            cls._init_collection(collection)
        return collection

    @classmethod
    def _get(cls, doc: dict):
        bson_id = doc['_id']
        if bson_id in cls._WEAKREFS:
            if obj := cls._WEAKREFS[bson_id]():
                # ensure the object isn't duplicated in memory, and uses any already existing one
                # this also ensures all references to the same db doc point to the same object
                return obj
        doc.update(cls._get_retrieve_vals(doc))
        return cls(**doc, ___internal___=True)

    @classmethod
    def _get_all_args(cls, *args, **kwargs):
        all_args = {k: v for k, v in zip(cls.__annotations__, args)}
        all_args.update(kwargs)
        return all_args

    @classmethod
    def query(cls, query) -> list[Self]:
        cls.lazy_init_subclass()
        return [cls._get(doc) for doc in cls.get_collection().find(query)]

    @classmethod
    def find(cls, *args, **kwargs) -> list[Self]:
        cls.lazy_init_subclass()
        all_args = cls._get_all_args(*args, **kwargs)
        return [cls._get(doc) for doc in cls.get_collection().find(cls._get_store_vals(all_args))]

    @classmethod
    def find_one(cls, *args, **kwargs) -> Optional[Self]:
        cls.lazy_init_subclass()
        all_args = cls._get_all_args(*args, **kwargs)
        if doc := cls.get_collection().find_one(cls._get_store_vals(all_args)):
            return cls._get(doc)

    @classmethod
    def delete_all(cls, *args, **kwargs):
        cls.lazy_init_subclass()
        all_args = cls._get_all_args(*args, **kwargs)
        cls.get_collection().delete_many(cls._get_store_vals(all_args))

    @classmethod
    def new(cls, *args, **kwargs):
        cls.lazy_init_subclass()
        all_args = cls._get_all_args(*args, **kwargs)
        doc = {k: all_args[k] if k in all_args else cls._get_default(k) for k in cls.__annotations__}
        try:
            insert_result = cls.get_collection().insert_one(cls._get_store_vals(doc))
        except pymongo.errors.DuplicateKeyError as e:
            raise UniqueConflict from e
        doc['_id'] = insert_result.inserted_id
        return cls(**doc, ___internal___=True)

    @classmethod
    def get_or_create(cls, *args, **kwargs):
        cls.lazy_init_subclass()
        return cls.find_one(*args, **kwargs) or cls(*args, **kwargs)

    @classmethod
    def _get_store_vals(cls, d: dict) -> dict:
        """Convert values in param d to database-storable values using cls._SERIALIZER

        :param d: dict containing raw values
        :return: dict containing database-storable values
        """
        return {k: cls._SERIALIZER.to_dict(v) for k, v in d.items()}

    @classmethod
    def _get_retrieve_vals(cls, d: dict) -> dict:
        """Convert values in param d from database-storable values back to their original values using cls._SERIALIZER
        :param d: dict containing database-storable values
        :return: dict containing raw values
        """
        rv = {}
        for k, v in d.items():
            if k in cls.__annotations__:
                rv[k] = cls._SERIALIZER.from_dict(cls.__annotations__[k], v)
            else:
                rv[k] = v
        return rv

    def update(self, **kwargs):
        """Update in-memory model instance with values in kwargs, and also write ONLY those changes to the database
        Behaviour differs from save method in that save also writes other changes:

        my_inst.field1 = 'not changed'
        my_inst.update(field2='changed too')  # ONLY writes field2 change to the database

        my_inst.field1 = 'not changed'
        my_inst.save(field2='changed too')  # Writes BOTH field1 and field2 change to the database
        """
        self.lazy_init_subclass()
        updates = {k: v for k, v in kwargs.items() if k in self.__annotations__}
        if updates:
            self.__dict__.update(updates)
            try:
                self.get_collection().update_one({'_id': self._id},
                                                 {'$set': self.__class__._get_store_vals(updates)})
            except pymongo.errors.DuplicateKeyError as e:
                raise UniqueConflict from e

    def delete(self):
        self.lazy_init_subclass()
        self.get_collection().delete_one({'_id': self._id})
        self._WEAKREFS.pop(self._id)

    def _db_store_(self):
        return getattr(self, self.__class__._KEY_FIELD)

    @classmethod
    def _db_retrieve_(cls, stored_val):
        return cls.find_one(**{cls._KEY_FIELD: stored_val})
